import numpy as np
import os
import epicure.epicuring as epi
import epicure.Utils as ut
import napari

def test_get_free_label():
    ## test from a skeletonized movie
    test_mov = os.path.join(".", "test_data", "003_crop.tif")
    test_seg = os.path.join(".", "test_data", "003_crop_epyseg.tif")
    epic = epi.EpiCure()
    epic.load_movie(test_mov)
    #epic.viewer = napari.Viewer(show=False)
    epic.go_epicure("epics", test_seg)
    assert len(epic.get_free_labels(5)) == 5
    assert len(epic.get_free_labels(15)) == 15
    epic = None

#def test_remove_layer():
#    """ Test remove layer utility """
#    test_mov = os.path.join(".", "test_data", "003_crop.tif")
#    test_seg = os.path.join(".", "test_data", "003_crop_epyseg.tif")
#    epic = epi.EpiCure()
#    epic.load_movie(test_mov)
#    epic.go_epicure("epics", test_seg)
#    assert "Events" in epic.viewer.layers 
#    ut.remove_layer(epic.viewer, "Events")
#    assert "Events" not in epic.viewer.layers 
#    epic.inspecting.first_event((0,10,20), 10, "suspect")
#    assert "Events" in epic.viewer.layers 
    
def utils_value():
    """ Labels operations in Utils """
    test_mov = os.path.join(".", "test_data", "003_crop.tif")
    test_seg = os.path.join(".", "test_data", "003_crop_epyseg.tif")
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    epic.load_movie(test_mov)
    epic.go_epicure("epics", test_seg)
    cell = epic.seg[0,160,80]
    assert cell > 0
    ut.changeLabel( epic.seglayer, cell, 30)
    cell = epic.seg[0,160,80]
    assert cell == 30

    ## read value under click 
    event = napari.utils.events.Event("mouse_press") 
    event.position = [0,160,80]
    event.view_direction = None
    event.dims_displayed = [0,1, 1] 
    cell = ut.getCellValue( epic.seglayer, event )
    assert cell == 30

    

if __name__ == "__main__":
    test_get_free_label()
    utils_value()
    print("********* Test operations cure completed ***********")


from napari import current_viewer
from magicgui import magicgui
from napari.utils.history import get_save_history, update_save_history 
import pathlib
import epicure.Utils as ut
from epicure.epicuring import EpiCure
import multiprocessing

"""
   Start EpiCure plugin
   Open the interface to select the movie and associated segmentation to process
"""

def start_from_layers():
    """ Start EpiCure from already opened image and segmentation layers """
    from typing import Union
    hist = get_save_history()
    cdir = hist[0]

    @magicgui(call_button="Use selected layers",)
    def select_layer(movie: Union["napari.layers.Image", None], movie_path: pathlib.Path, segmentation: Union["napari.layers.Layer", None]):
        """ GUI to choose the layers to use """
        if movie == "None":
            movie = None
        segmented = segmentation
        if segmentation == "None":
            segmented = None
        ut.remove_all_widgets(viewer)
        #ut.remove_widget(viewer, "Start from opened layers")
        start_epi, epicure_instance = gui_files(movie=movie, movie_path=movie_path, segmented=segmented)
        viewer.window.add_dock_widget(start_epi)
        return start_epi

    viewer = current_viewer()
    wid = viewer.window.add_dock_widget(select_layer)
    return wid


def start_epicure():
    """ Start EpiCure from scratch """
    gui, epicure = gui_files(movie=None, movie_path ="", segmented=None)
    return gui

def gui_files(movie=None, movie_path="", segmented=None):
    """ GUI to choose files and parameters """
    hist = get_save_history()
    cdir = hist[0]
    viewer = current_viewer()
    Epic = EpiCure(viewer)
    caxis = None
    cval = 0
    ncpus = int(multiprocessing.cpu_count()*0.5)

    def set_visibility():
        """ Visibility of the parameters in the GUI """
        if movie is not None:
            get_files.image_file.visible = False
        if segmented is not None:
            get_files.segmentation_file.visible = False
            get_files.segment_with_epyseg.visible = False
        advanced_visibility()

    def advanced_visibility():
        """ Handle the visibility of the advanced parameters """
        get_files.output_dirname.visible = get_files.advanced_parameters.value
        get_files.show_other_chanels.visible = get_files.advanced_parameters.value
        get_files.process_frames_parallel.visible = get_files.advanced_parameters.value
        get_files.nbparallel_threads.visible = get_files.advanced_parameters.value
        get_files.junction_half_thickness.visible = get_files.advanced_parameters.value
        get_files.verbose_level.visible = get_files.advanced_parameters.value
        get_files.allow_gaps.visible = get_files.advanced_parameters.value
        get_files.show_scale_bar.visible = get_files.advanced_parameters.value
        get_files.epithelial_cells.visible = get_files.advanced_parameters.value

    def load_movie_from_layers(movie, movie_path, segmented):
        """ Load and display the selected layers """
        start_time = ut.start_time()
        nonlocal caxis, cval
        caxis, cval = Epic.movie_from_layer(movie, movie_path)
        imgdir = ut.get_directory(movie_path)
        
        if segmented is None:
            get_files.segmentation_file.visible = True
            get_files.segmentation_file.value = pathlib.Path(imgdir)
            get_files.segment_with_epyseg.visible = True
        
        labname = Epic.suggest_segfile( get_files.output_dirname.value )
        Epic.set_names( get_files.output_dirname.value )
        if labname is not None:
            get_files.segmentation_file.value = pathlib.Path(labname)
            Epic.read_epicure_metadata()    
        if caxis is not None:
            get_files.junction_chanel.max = cval-1
            get_files.junction_chanel.visible = True
            set_chanel()
        show_metatdata(show=True)
        get_files.allow_gaps.value = bool(Epic.epi_metadata["Allow gaps"])
        get_files.verbose_level.value = int(Epic.epi_metadata["Verbose"])
        get_files.call_button.enabled = True

    def show_metatdata(show=True):
        """ Show or update the metadata parameters """
        get_files.scale_xy.value = Epic.epi_metadata["ScaleXY"]
        get_files.timeframe.value = Epic.epi_metadata["ScaleT"]
        get_files.unit_xy.value = Epic.epi_metadata["UnitXY"]
        get_files.unit_t.value = Epic.epi_metadata["UnitT"]
        get_files.scale_xy.visible = show 
        get_files.unit_xy.visible = show
        get_files.timeframe.visible = show 
        get_files.unit_t.visible = show 

    def load_movie():
        """ Load and display the selected movie """
        start_time = ut.start_time()
        nonlocal caxis, cval
        image_file = get_files.image_file.value
        caxis, cval = Epic.load_movie(image_file)
        imgdir = ut.get_directory(image_file)
        get_files.segmentation_file.visible = True
        get_files.segmentation_file.value = pathlib.Path(imgdir)
        labname = Epic.suggest_segfile( get_files.output_dirname.value )
        Epic.set_names( get_files.output_dirname.value )
        if labname is not None:
            get_files.segmentation_file.value = pathlib.Path(labname)
            Epic.read_epicure_metadata()    
        if caxis is not None:
            get_files.junction_chanel.max = cval-1
            get_files.junction_chanel.visible = True
            set_chanel()
        show_metatdata(show=True)
        get_files.segment_with_epyseg.visible = True
        get_files.allow_gaps.value = bool(Epic.epi_metadata["Allow gaps"])
        get_files.verbose_level.value = int(Epic.epi_metadata["Verbose"])
        get_files.call_button.enabled = True
        ut.show_duration(start_time, header="Movie loaded in ")

    def show_others():
        """ Display other chanels from the initial movie """
        for ochan in range(cval):
            ut.remove_layer(viewer, "MovieChannel_"+str(ochan))
        if get_files.show_other_chanels.value == True:
            Epic.add_other_chanels(int(get_files.junction_chanel.value), caxis)

    def set_chanel():
        """ Set the correct chanel that contains the junction signal """
        start_time = ut.start_time()
        Epic.set_chanel( int(get_files.junction_chanel.value), caxis )
        show_others()
        ut.show_duration(start_time, header="Movie chanel loaded in ")

    def launch_napari_epyseg():
        """ Open napari-epyseg plugin to segment the intensity channel movie """
        try:
            import napari_epyseg
            from napari_epyseg.start_epyseg import run_epyseg
        except:
            ut.show_error( "This option requires the plugin napari-epyseg that is missing.\nInstall it and restart" )
            return
        print("Running EpySeg with default parameters on the movie. To change the settings, use the napari-epyseg plugin outside of EpiCure or EpySeg module directly")
        parameters = {"tile_width":256, "tile_height":256, "overlap_width":32, "overlap_height":32, "model":"epyseg default(v2)", "norm_min":0, "norm_max":1}
        print(Epic.img.shape)
        ut.show_progress( viewer, True )
        segres = run_epyseg( Epic.img, parameters, prog=True )
        ut.show_progress( viewer, False )
        segname = str(get_files.image_file.value)+"_epyseg.tif"
        ut.writeTif( segres, segname, 1.0, "uint8", what="Epyseg results saved in " )
        get_files.segmentation_file.value = segname
        get_files.segment_with_epyseg.visible = False


    @magicgui(call_button="START CURE",
            junction_chanel={"widget_type": "Slider", "min":0, "max": 0},
            _ = {"widget_type": "Label"},
            scale_xy = {"widget_type": "LiteralEvalLineEdit"},
            timeframe = {"widget_type": "LiteralEvalLineEdit"},
            __ = {"widget_type": "Label"},
            _____ = {"widget_type": "Label"},
            ______ = {"widget_type": "Label"},
            segment_with_epyseg = {"widget_type": "PushButton", "label": "Segment now with EpySeg"},
            ___ = {"widget_type": "Label"},
            junction_half_thickness={"widget_type": "LiteralEvalLineEdit"},
            nbparallel_threads = {"widget_type": "LiteralEvalLineEdit"},
            verbose_level={"widget_type": "Slider", "min":0, "max": 3},
            )
    def get_files( 
                   image_file = pathlib.Path(cdir),
                   junction_chanel = 0,
                   _ = "Image metadata",
                   scale_xy = 1,
                   unit_xy = "um",
                   timeframe = 1,
                   unit_t = "min",
                   __ = "\nSegmentation\n",
                   _____ = "Load segmentation or TrackMateXML file",
                   segmentation_file = pathlib.Path(cdir),
                   ______ = "OR\t",
                   segment_with_epyseg = False,
                   ___ = "",
                   advanced_parameters = False,
                   show_other_chanels = True,
                   show_scale_bar = True,
                   allow_gaps = True,
                   epithelial_cells = True,
                   process_frames_parallel = False,
                   nbparallel_threads = ncpus,
                   junction_half_thickness = 1,
                   output_dirname = "epics",
                   verbose_level = 1,
                   ):

        print("Starting")
        imname, imdir, outdir = ut.extract_names( image_file, output_dirname )
        update_save_history(imdir)
        #ut.remove_widget(viewer, "Start EpiCure (epicure)")
        ut.remove_all_widgets( viewer )
        Epic.process_parallel = process_frames_parallel
        Epic.set_verbose( verbose_level )
        Epic.nparallel = nbparallel_threads
        #Epic.load_segmentation(segmentation_file)
        #Epic.check_shape()
        Epic.set_thickness( junction_half_thickness )
        Epic.set_scales(scale_xy, timeframe, unit_xy, unit_t)
        Epic.set_scalebar( show_scale_bar )
        Epic.set_gaps_option( allow_gaps )
        Epic.set_epithelia( epithelial_cells )
        ## to handle segmentation from layer or from file
        segmentation_input = {"File":segmentation_file} 
        if segmented is not None:
            segmentation_input["Layer"]=segmented

        Epic.go_epicure(outdir, segmentation_input)
    
    set_visibility()
    get_files.call_button.enabled = False
    get_files.segmentation_file.visible = False
    get_files.segment_with_epyseg.visible = False
    get_files.scale_xy.visible = False
    get_files.unit_xy.visible = False
    get_files.timeframe.visible = False
    get_files.unit_t.visible = False
    if movie is not None:
        load_movie_from_layers(movie, movie_path, segmented)
        get_files.image_file.value = movie_path
        show_metatdata(show=True)
        if segmented is not None:
            get_files.call_button.enabled = True
    get_files.junction_chanel.visible = False
    get_files.advanced_parameters.clicked.connect(set_visibility)
    get_files.show_other_chanels.clicked.connect(show_others)
    get_files.image_file.changed.connect(load_movie)
    get_files.junction_chanel.changed.connect(set_chanel)
    get_files.segment_with_epyseg.clicked.connect( launch_napari_epyseg )
    return get_files, Epic




from datetime import datetime
from pathlib import Path
from typing import Dict, List
from xml.dom import minidom
import xml.etree.ElementTree as ET
#import cv2
#import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.cluster.hierarchy import DisjointSet
from skimage.measure import find_contours

import epicure.Utils as ut


# TODO: deal with groups (store in MANUAL_SPOT_COLOR)

SPOT_FEATS = [
    {"feature": "POSITION_X", "name": "X", "shortname": "X", "dimension": "POSITION", "isint": "false"},
    {"feature": "POSITION_Y", "name": "Y", "shortname": "Y", "dimension": "POSITION", "isint": "false"},
    {"feature": "POSITION_Z", "name": "Z", "shortname": "Z", "dimension": "POSITION", "isint": "false"},
    {"feature": "POSITION_T", "name": "T", "shortname": "T", "dimension": "TIME", "isint": "false"},
    {"feature": "FRAME", "name": "Frame", "shortname": "Frame", "dimension": "NONE", "isint": "true"},
    {"feature": "VISIBILITY", "name": "Visibility", "shortname": "Visibility", "dimension": "NONE", "isint": "true"},
    {"feature": "RADIUS", "name": "Radius", "shortname": "R", "dimension": "LENGTH", "isint": "false"},
    {"feature": "MANUAL_SPOT_COLOR", "name": "Manual spot color", "shortname": "Spot color", "dimension": "NONE", "isint": "true"},
]

EDGE_FEATS = [
    {"feature": "SPOT_SOURCE_ID", "name": "Source spot ID", "shortname": "Source ID", "dimension": "NONE", "isint": "true"},
    {"feature": "SPOT_TARGET_ID", "name": "Target spot ID", "shortname": "Target ID", "dimension": "NONE", "isint": "true"},
]

TRACK_FEATS = [
    {"feature": "TRACK_INDEX", "name": "Track index", "shortname": "Index", "dimension": "NONE", "isint": "true"},
    {"feature": "TRACK_ID", "name": "Track ID", "shortname": "ID", "dimension": "NONE", "isint": "true"},
]


def build_feat_declaration_tag():
    """Build the FeatureDeclarations tag for TrackMate XML."""
    feat_declarations = ET.Element("FeatureDeclarations")
    spot_feats = ET.SubElement(feat_declarations, "SpotFeatures")
    for feat in SPOT_FEATS:
        ET.SubElement(spot_feats, "Feature", feat)
    edge_feats = ET.SubElement(feat_declarations, "EdgeFeatures")
    for feat in EDGE_FEATS:
        ET.SubElement(edge_feats, "Feature", feat)
    track_feats = ET.SubElement(feat_declarations, "TrackFeatures")
    for feat in TRACK_FEATS:
        ET.SubElement(track_feats, "Feature", feat)
    return feat_declarations


def build_spots_df(epic):
    """Build a DataFrame representing the spots table for TrackMate XML."""
    df_spots = pd.DataFrame(epic.tracking.track_data, columns=["label", "FRAME", "pos_x", "pos_y"])
    df_spots["ID"] = df_spots.index
    df_spots["name"] = df_spots.apply(lambda row: f"LABEL{row['label']}_FRAME{row['FRAME']}", axis=1)
    # Invert X and Y to match TrackMate convention.
    df_spots["POSITION_X"] = df_spots["pos_y"] * epic.epi_metadata.get("ScaleXY", 1)
    df_spots["POSITION_Y"] = df_spots["pos_x"] * epic.epi_metadata.get("ScaleXY", 1)
    df_spots["POSITION_Z"] = 0.0  # 2D data. 
    df_spots["POSITION_T"] = df_spots["FRAME"] * epic.epi_metadata.get("ScaleT", 1)
    df_spots["VISIBILITY"] = 1
    df_spots.drop(columns=["pos_x", "pos_y"], inplace=True)

    return df_spots


def get_cell_contour(seg_array, label, frame):
    """Get the contours of the cell with the given label in the given frame."""
    spot_seg = seg_array[frame, :, :] == label
    # print(spot_seg.dtype)
    # Pad to avoid contours on the border (skimage.measure.find_contours)
    spot_seg = np.pad(spot_seg, pad_width=1, mode='constant', constant_values=0)
    #spot_seg_uint8 = spot_seg.astype(np.uint8)
    # print(spot_seg_uint8.dtype)

    if np.sum(spot_seg) > 0:
        contours = find_contours(spot_seg, level=0.5)
        #contours, _ = cv2.findContours(spot_seg_uint8, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        # print(type(contours))
        # if label == 125 and frame == 0:
        #     contours = contours[:-1]  # hack to remove spurious contour

        # plt.show()
        # print(contours)
        assert len(contours) > 0, f"No contour found for label {label} in frame {frame}"
        # assert len(contours) < 2, f"More than one contour found for label {label} in frame {frame}"
        if len(contours) > 1:
             print(f"Warning, found two cells for label {label} at frame {frame}")
             print(f"Keep only the first one in the export")
        #if label==6:
        #     fig, ax = plt.subplots(1, 1)
        #     ax.imshow(seg_array[frame,:,:]>0)
             #ax.imshow(spot_seg, cmap="grey")
             # Plot each contour
        #     for contour in contours:
        #         contour = contour.squeeze()
        #         ax.plot(contour[:, 1], contour[:, 0], linewidth=1)

        #     plt.show()

        return np.flip(contours[0].squeeze(), axis=1)
    return None


def build_roi_contours(epic, df_spots):
    """Build a mapping from spot ID to its ROI contour."""
    roi_contours = {}
    seg_layer = epic.seg
    for _, spot in df_spots.iterrows():
        contour = get_cell_contour(seg_layer, spot["label"], spot["FRAME"])
        if contour is not None:
            # Convert from pixels to space units.
            contour = contour.astype(np.float64)  # in case the scale is a float
            # Remove the padding
            contour[:, 0] -= 1 
            contour[:, 1] -= 1 
            contour[:, 0] *= epic.epi_metadata.get("ScaleXY", 1)
            contour[:, 1] *= epic.epi_metadata.get("ScaleXY", 1)
            # Convert contour from absolute to relative coordinates (to the cell XY position).
            contour[:, 0] -= spot["POSITION_X"]
            contour[:, 1] -= spot["POSITION_Y"]
            # Flatten contour to a 1D array as expected by TrackMate.
            contour = contour.flatten()
            roi_contours[spot["ID"]] = contour
            # print(spot["ID"], spot["label"], spot["FRAME"], contour.shape)
    return roi_contours


def build_all_spots_tag(df_spots, roi_n_points):
    """Build the AllSpots tag for TrackMate XML."""
    all_spots = ET.Element("AllSpots", {"nspots": str(len(df_spots))})
    frames = df_spots["FRAME"].unique()
    # print(roi_n_points)
    for frame in frames:
        spots_in_frame = df_spots[df_spots["FRAME"] == frame]
        frame_tag = ET.SubElement(all_spots, "SpotsInFrame", {"frame": str(frame)})
        for _, spot in spots_in_frame.iterrows():
            spot_attrib = {
                "ID": str(spot["ID"]),
                "name": spot["name"],
                "EpiCure_label": str(spot["label"]),
                "POSITION_X": str(spot["POSITION_X"]),
                "POSITION_Y": str(spot["POSITION_Y"]),
                "POSITION_Z": str(spot["POSITION_Z"]),
                "POSITION_T": str(spot["POSITION_T"]),
                "FRAME": str(spot["FRAME"]),
                "VISIBILITY": str(spot["VISIBILITY"]),
                "RADIUS": "-1",
                "MANUAL_SPOT_COLOR": "0",  # TODO: use it to store group
            }

            # Add ROI contour if available.
            if spot["ID"] in roi_n_points:
                spot_attrib["ROI_N_POINTS"] = str(roi_n_points[spot["ID"]].shape[0] // 2)

            spot_tag = ET.SubElement(frame_tag, "Spot", spot_attrib)
            if spot["ID"] in roi_n_points:
                spot_tag.text = " ".join(map(str, roi_n_points[spot["ID"]]))

    return all_spots


def create_label_to_track_mapping(divisions: Dict[int, List[int]], unique_labels: List[int]) -> Dict[int, int]:
    """
    Create a mapping from labels to track IDs using scipy's DisjointSet for efficient track grouping.

    Args:
        divisions: dict of {daughter_label: [mother_labels]} from epic.tracking.graph
        unique_labels: list of unique labels present in the tracking data

    Returns:
        dict: {label: track_id} - mapping from each label to its track ID
    """
    if not divisions:
        # No divisions - each unique label is its own track.
        return {label: label for label in unique_labels}

    ds = DisjointSet(unique_labels)

    # Union connected labels based on mother-daughter relationships.
    for daughter, mothers in divisions.items():
        if daughter not in unique_labels:  # weirdly, this can happen
            continue
        for mother in mothers:
            if mother in unique_labels:
                ds.merge(daughter, mother)

    # A connected component is a TrackMate track. We use the root as track ID.
    # Create a mapping from label to track_id (root).
    label_to_track_id = {}
    for label in unique_labels:
        root = ds[label]
        label_to_track_id[label] = root

    return label_to_track_id


def build_all_tracks_data(divisions, df_spots):
    """"""
    print(f"Divisions: {divisions}")
    for mothers in divisions.values():
        if len(mothers) > 1:
            print("FUSION")

    # Generate and assign TRACK_IDs.
    labels = df_spots["label"].unique()
    label_to_track_id = create_label_to_track_mapping(divisions, labels)
    df_spots["TRACK_ID"] = df_spots["label"].map(label_to_track_id)

    # Division edges: for each daughter-mother pair, create an edge.
    edges_data = [{"daughter": daughter, "mother": mother} for daughter, mothers in divisions.items() for mother in mothers]
    df_edges = pd.DataFrame(edges_data)
    # Labels stay the same until there is a division. But spots ID are unique.
    # It means that in df_spots, labels appears multiple times. Because of this
    # we cannot easily map between df_spots and df_edges. So we create intermediary
    # columns to ease the mapping.
    df_spots["first_frame"] = df_spots.groupby("label")["FRAME"].transform("min")
    df_spots["last_frame"] = df_spots.groupby("label")["FRAME"].transform("max")
    # A daughter is at the first frame of its label, a mother at the last frame of its label.
    df_spots["daughter"] = df_spots["first_frame"] == df_spots["FRAME"]
    df_spots["mother"] = df_spots["last_frame"] == df_spots["FRAME"]
    df_spots.drop(columns=["first_frame", "last_frame"], inplace=True)
    # Now we can map between df_spots and df_edges.
    # The SPOT_SOURCE_ID is the spot ID of the matching label that is a mother,
    # and the SPOT_TARGET_ID is the spot ID of the matching label that is a daughter.
    df_edges["SPOT_SOURCE_ID"] = df_edges["mother"].map(df_spots[df_spots["mother"]].set_index("label")["ID"])
    df_edges["SPOT_TARGET_ID"] = df_edges["daughter"].map(df_spots[df_spots["daughter"]].set_index("label")["ID"])
    df_spots.drop(columns=["daughter", "mother"], inplace=True)

    # Add TRACK_ID to division edges (needed for XML).
    if not df_edges.empty:
        df_edges["TRACK_ID"] = df_edges["mother"].map(label_to_track_id)
        df_edges.drop(columns=["daughter", "mother"], inplace=True)

    # Non-division edges: for each label, connect consecutive spots within that label.
    non_division_edges = []
    for label in df_spots["label"].unique():
        label_spots = df_spots[df_spots["label"] == label].sort_values("FRAME")
        if len(label_spots) > 1:
            track_id = label_spots.iloc[0]["TRACK_ID"]
            for i in range(len(label_spots) - 1):
                current_spot = label_spots.iloc[i]
                next_spot = label_spots.iloc[i + 1]
                non_division_edges.append({"SPOT_SOURCE_ID": current_spot["ID"], "SPOT_TARGET_ID": next_spot["ID"], "TRACK_ID": track_id})

    # Combine division and non-division edges.
    df_non_division_edges = pd.DataFrame(non_division_edges)
    if not df_edges.empty and not df_non_division_edges.empty:
        # Make sure both dataframes have the same columns.
        df_edges = df_edges[["SPOT_SOURCE_ID", "SPOT_TARGET_ID", "TRACK_ID"]]
        df_edges = pd.concat([df_edges, df_non_division_edges], ignore_index=True)
    elif not df_non_division_edges.empty:
        df_edges = df_non_division_edges

    # Final cleanup and type conversion.
    if not df_edges.empty:
        # We can have NaN if a label has no mother (appears at first frame)
        # or no daughter (disappears at last frame).
        df_edges.dropna(inplace=True)
        # Convert to int in case of NaN.
        df_edges["SPOT_SOURCE_ID"] = df_edges["SPOT_SOURCE_ID"].astype(int)
        df_edges["SPOT_TARGET_ID"] = df_edges["SPOT_TARGET_ID"].astype(int)
        df_edges["TRACK_ID"] = df_edges["TRACK_ID"].astype(int)

    return df_edges


def build_all_tracks_tag(df_edges):
    """Build the AllTracks tag for TrackMate XML."""
    all_tracks = ET.Element("AllTracks")
    track_ids = sorted(df_edges["TRACK_ID"].unique())
    track_id_to_index = {track_id: index for index, track_id in enumerate(track_ids)}

    for track_id in track_ids:
        track = ET.SubElement(all_tracks, "Track")
        track.set("name", f"Track_{track_id}")
        track.set("TRACK_ID", str(track_id))
        track.set("TRACK_INDEX", str(track_id_to_index[track_id]))

        track_edges = df_edges[df_edges["TRACK_ID"] == track_id]
        for _, edge in track_edges.iterrows():
            edge_attrib = {
                "SPOT_SOURCE_ID": str(edge["SPOT_SOURCE_ID"]),
                "SPOT_TARGET_ID": str(edge["SPOT_TARGET_ID"]),
            }
            ET.SubElement(track, "Edge", edge_attrib)

    return all_tracks


def build_filtered_tracks_tag(track_ids):
    """Build the FilteredTracks tag for TrackMate XML."""
    filtered_tracks = ET.Element("FilteredTracks")
    for track_id in track_ids:
        ET.SubElement(filtered_tracks, "TrackID", {"TRACK_ID": str(track_id)})
    return filtered_tracks


def build_model_tag(epic):
    """Build the Model tag for TrackMate XML."""
    model = ET.Element("Model")
    model.set("spatialunits", epic.epi_metadata.get("UnitXY", "pixel"))
    model.set("timeunits", epic.epi_metadata.get("UnitT", "frame"))
    model.append(build_feat_declaration_tag())

    print("Tracked?", epic.tracked)
    df_spots = build_spots_df(epic)
    cell_contours = build_roi_contours(epic, df_spots)
    model.append(build_all_spots_tag(df_spots, cell_contours))
    divisions = epic.tracking.graph  # dict of {daughter: mothers}
    if divisions:
        df_edges = build_all_tracks_data(divisions, df_spots)
        model.append(build_all_tracks_tag(df_edges))
        track_ids = sorted(df_edges["TRACK_ID"].unique())
        model.append(build_filtered_tracks_tag(track_ids))
    else:
        model.append(ET.Element("AllTracks"))
        model.append(ET.Element("FilteredTracks"))

    return model


def build_settings_tag(epic):
    """Build the Settings tag for TrackMate XML."""
    settings = ET.Element("Settings")
    img_path = Path(epic.epi_metadata["MovieFile"])
    ET.SubElement(
        settings,
        "ImageData",
        {
            "filename": img_path.name,
            "folder": str(img_path.parent),  # TODO: missing / at the end compared to TM, is it an issue?
            "width": str(epic.imgshape2D[1]),
            "height": str(epic.imgshape2D[0]),
            "nslices": "1",
            "nframes": str(epic.nframes),
            "pixelwidth": str(epic.epi_metadata.get("ScaleXY", 1)),
            "pixelheight": str(epic.epi_metadata.get("ScaleXY", 1)),
            "voxeldepth": "1",
            "timeinterval": str(epic.epi_metadata.get("ScaleT", 1)),
        },
    )
    ET.SubElement(
        settings,
        "BasicSettings",
        {
            "xstart": "0",
            "xend": str(epic.imgshape2D[1] - 1),
            "ystart": "0",
            "yend": str(epic.imgshape2D[0] - 1),
            "zstart": "0",
            "zend": "0",
            "tstart": "0",
            "tend": str(epic.nframes - 1),
        },
    )
    ET.SubElement(settings, "DetectorSettings", {"DETECTOR_NAME": "MANUAL_DETECTOR", "RADIUS": "5"})
    ET.SubElement(settings, "InitialSpotFilter")
    ET.SubElement(settings, "SpotFilterCollection")
    ET.SubElement(settings, "TrackerSettings", {"TRACKER_NAME": "MANUAL_TRACKER"})
    ET.SubElement(settings, "TrackFilterCollection")
    return settings


# def build_gui_state_tag():
#     """Build the GUIState tag for TrackMate XML."""
#     gui_state = ET.Element("GUIState")
#     return gui_state


def pretty_print_xml(element):
    """Pretty print an XML element."""
    rough_string = ET.tostring(element, encoding="utf-8")
    parsed = minidom.parseString(rough_string)
    return parsed.toprettyxml(indent="  ")


def save_trackmate_xml(epic, outname):
    """Save a TrackMate XML file."""
    if epic.verbose == 3:
        ut.show_info(f"ScaleXY: {epic.epi_metadata.get('ScaleXY')}")
        ut.show_info(f"ScaleT: {epic.epi_metadata.get('ScaleT')}")
        ut.show_info(f"imgshape2D: {epic.imgshape2D}")
        ut.show_info(f"UnitXY: {epic.epi_metadata.get('UnitXY')}")
        ut.show_info(f"UnitT: {epic.epi_metadata.get('UnitT')}")

    root = ET.Element("TrackMate", {"version": "unknown"})
    log = ET.SubElement(root, "Log")
    now = datetime.now()
    log.text = f"Created by EpiCure on {now.strftime('%Y-%m-%d %H:%M:%S')}"
    model = build_model_tag(epic)
    root.append(model)
    settings = build_settings_tag(epic)
    root.append(settings)
    ET.SubElement(root, "GUIState", {"state": "ConfigureViews"})
    ET.SubElement(root, "DisplaySettings")
    # display_settings = ET.SubElement(root, "DisplaySettings")
    # display_settings.text = "{'spotDisplayedAsRoi': true}"
    # tree = ET.ElementTree(root)

    pretty_xml = pretty_print_xml(root)
    # TODO: check if ET.indent is better or more efficient than pretty_print_xml
    # for Python 3.9+
    # Pretty formatting - ET.indent is available from Python 3.9+
    # try:
    #     ET.indent(tree, space="  ", level=0)
    # except AttributeError:
    #     # Fallback for Python < 3.9
    #     pass
    with open(outname, "w", encoding="utf-8") as f:
        f.write(pretty_xml)



import numpy as np
import os
import epicure.epicuring as epi
import epicure.Utils as ut
#from unittest.mock import Mock
#from vispy import keys
import napari


def test_output_selected():
    """ Selection of cells to export segmentation as ROI/labels """
    test_img = os.path.join(".", "test_data", "area3_Composite.tif")
    test_seg = os.path.join(".", "test_data", "area3_Composite_epyseg.tif")
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    resaxis, resval = epic.load_movie(test_img)
    epic.set_chanel(1, 1)
    assert epic.viewer is not None
    epic.go_epicure("test_epics", test_seg)

    ## test changing the current selection of export (one cell or all cells)
    output = epic.outputing
    assert output is not None
    output.output_mode.setCurrentText("All cells")
    sel = output.get_selection_name()
    assert sel == ""
    output.output_mode.setCurrentText("Only selected cell")
    sel = output.get_selection_name()
    assert sel == "_cell_1"
    ## export a cell segmentation as Fiji ROI
    output.save_choice.setCurrentText("ROI")
    roi_file = os.path.join(".", "test_data", "test_epics", "area3_Composite_rois_cell_1.zip")
    if os.path.exists(roi_file):
        os.remove(roi_file)
    output.save_segmentation()
    assert os.path.exists(roi_file)
    
def test_measure_vertices():
    """ Find the vertices (TCJ) and measure their intensity and nb of neighbors """

    test_img = os.path.join(".", "test_data", "area3_Composite.tif")
    test_seg = os.path.join(".", "test_data", "area3_Composite_epyseg.tif")
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    resaxis, resval = epic.load_movie(test_img)
    epic.set_chanel(1, 1)
    assert epic.viewer is not None
    epic.go_epicure("test_epics", test_seg)

    output = epic.outputing
    assert output is not None

    # Default vertex sizes
    output.vertice_radius.setText("1.25")
    output.measure_vertices()
    assert "Vertices" in viewer.layers
    vertices = viewer.layers["Vertices"].data
    frame = 1
    props = ut.binary_properties(vertices[frame])
    nvertex = len(props)
    assert nvertex > 150
    assert nvertex < 250
    nneigh = np.array((output.table["nb_neighbors"]))
    ## Nb of neighbor shuold in majority 3, and between 2 and 6-ish max
    assert np.min(nneigh>=2)
    assert np.max(nneigh<=6)
    assert np.floor(np.median(nneigh)) == 3

    ## Take bigger vertex: merge neighboring ones
    output.vertice_radius.setText("4")
    output.measure_vertices()
    vertices = viewer.layers["Vertices"].data
    props = ut.binary_properties(vertices[frame])
    assert len(props)>100
    assert len(props)<nvertex

def test_measure_intensities():
    """ Test measure of cell features, in particular intensity measurement """

    test_img = os.path.join(".", "test_data", "area3_Composite.tif")
    test_seg = os.path.join(".", "test_data", "area3_Composite_epyseg.tif")
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    resaxis, resval = epic.load_movie(test_img)
    epic.set_chanel(1, 1)
    # test when there are two channels to measure
    epic.add_other_chanels(0, 1)  ## keep also the first channel to measure intensity in it
    assert epic.viewer is not None
    epic.go_epicure("test_epics", test_seg)

    output = epic.outputing
    assert output is not None
    output.output_mode.setCurrentText("All cells")

    ## Make sure that intensity_junction_cytoplasm measure is selected
    output.cell_features.select_all("intensity")

    ## select the two channels to measure
    output.cell_features.chan_list.item(0).setSelected(True)
    output.cell_features.chan_list.item(1).setSelected(True)

    ## go measure the cell features
    output.measure_features()

    ## check that the intensity measurements were done in both channels
    assert "intensity_mean_MovieChannel_1" in output.table.keys()
    assert "intensity_mean" in output.table.keys()
    assert "intensity_junction_MovieChannel_1" in output.table.keys()

    ## now remove the other channels to check it runs on only one channel
    epic.others = None
    output.cell_features.chan_list = None
    ## go measure the cell features
    output.measure_features()
    ## check that it measured only the movie channel
    assert "intensity_mean" in output.table.keys()
    assert "intensity_mean_MovieChannel_1" not in output.table.keys()

def test_measure_events():
    """ Measure/export of events """
    test_img = os.path.join(".", "test_data", "013_crop.tif")
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    epic.load_movie(test_img)
    epic.go_epicure("epics")
    output = epic.outputing

    ## Export of number of events by frame
    output.output_mode.setCurrentText("All cells") ## ensure to measure all cells
    tab = output.count_events()
    ## should have only one division at frame 3
    assert int(tab.loc[tab["frame"]==2, 'division'].iloc[0]) == 0
    assert int(tab.loc[tab["frame"]==3, 'division'].iloc[0]) == 1
    assert np.sum(tab['extrusion']) == 0

if __name__ == "__main__":
    #test_output_selected()
    #test_measure_events()
    test_measure_vertices()
    print("********* Test outputs cure completed ***********")



"""
Test the "export to trackmate XML" options in Output. 
Corresponding code in src/epicure/trackmate_export.py
"""

import epicure.epicuring as epicure
import os

def test_export_trackmate():
    """ Export tracks as a TrackMate XML file """
    raw_path = os.path.join( ".", "test_data", "013_crop.tif" )
    xml_path = os.path.join(".", "test_data", "epics", "013_crop.xml")
    if os.path.exists(xml_path):
        os.remove(xml_path)

    epic = epicure.EpiCure()
    epic.verbose = 3  # 0: minimal to 3: debug informations
    epic.load_movie(raw_path)
    epic.go_epicure(outdir="epics")

    epic.outputing.output_mode.setCurrentText("All cells")
    epic.outputing.save_tm_xml()

    ## check that metadata were well read
    assert epic.epi_metadata["ScaleXY"]==0.2
    assert int(epic.epi_metadata["ScaleT"])==300 
    ## check that TrackMate file was generated
    assert os.path.exists(xml_path)

def test_tm_loader_trackmate():
    """ Test functions to import a Trackmate .xml tracks """
    import epicure.tm_loader as tm
    import numpy as np
    from pathlib import Path
    tm_file = "test_data/FakeTracks_with_fusions.xml"
    # tm_file = "/media/lxenard/data/Code/pycellin/pycellin/sample_data/Ecoli_growth_on_agar_pad.xml"
    # tm_file = "/home/lxenard/snap/trackMateMoiCa/epics/013_crop.xml"
    np.set_printoptions(suppress=True, floatmode="maxprec_equal")

    img_data_tag = tm._get_ImageData_tag(Path(tm_file))
    metadata = tm._get_metadata(img_data_tag)
    seg_shape = (int(metadata["nframes"]), int(metadata["height"]), int(metadata["width"]))
    segmentation = np.zeros(seg_shape, dtype=np.uint16)
    positions, tracks = tm._parse_Model_tag(Path(tm_file), metadata, segmentation)
    label_mapping = tm._build_label_mapping(positions, tracks)
    positions = tm.relabel_positions(label_mapping, positions)
    tracks = tm.relabel_tracks(label_mapping, tracks)
    segmentation = tm.relabel_segmentation(label_mapping, segmentation)

    # import matplotlib.pyplot as plt

    # n_frames = segmentation.shape[0]
    # n_cols = 10
    # n_rows = (n_frames + n_cols - 1) // n_cols
    # fig, axes = plt.subplots(n_rows, n_cols, figsize=(40, 2 * n_rows))
    # for i in range(n_frames):
    #     ax = axes[i // n_cols, i % n_cols]
    #     ax.imshow(segmentation[i], cmap="gray")
    #     ax.set_title(f"Frame {i}")
    #     ax.axis("off")
    # plt.tight_layout()
    # plt.show()

    ## Test relabeling
    # Expected mapping from TrackMate labels to EpiCure labels for the test file.
    expected_tracklets = {
        "a": [2004, 2005, 2007, 2009, 2010, 2011, 2013, 2014, 2015],
        "b": [2006, 2008],
        "c": [2012],
        "d": [2017, 2020, 2021, 2024, 2026, 2028, 2029, 2031, 2034, 2038, 2042, 2045, 2047, 2051, 2053, 2056, 2059, 2064],
        "e": [2016, 2019, 2022, 2023, 2025, 2027, 2030],
        "f": [2018],
        "g": [2033, 2036, 2039, 2043, 2044, 2049, 2052, 2054, 2058, 2060, 2062, 2065, 2068, 2070],
        "h": [2032, 2035, 2040, 2041, 2046, 2048, 2050, 2055, 2057, 2061, 2063],
        "i": [2037],
        "j": [2066, 2067, 2071, 2072, 2074, 2076, 2078, 2080, 2083, 2084, 2085, 2088, 2090, 2092, 2095],
        "k": [2069],
        "l": [2073],
        "m": [2075, 2077, 2079, 2081],
        "n": [2082],
        "o": [2086],
        "p": [2087, 2089],
        "q": [2091, 2093, 2096, 2097, 2099, 2100, 2102, 2103, 2105, 2108],
        "r": [2094],
        "s": [2098],
        "t": [2101],
        "u": [2106, 2109],
        "v": [2107, 2110],
    }

    # Test that we have the same number of tracklets as expected.
    nb_tracklets_exp = len(expected_tracklets)
    nb_tracklets_obt = len(set(positions[:, 0]))
    assert nb_tracklets_exp == nb_tracklets_obt, f"Expected {nb_tracklets_exp} tracklets, but got {nb_tracklets_obt}."

    # Test that detections get grouped into tracklets as expected
    # Build a mapping from new labels to sets of original TrackMate IDs
    obtained_tracklets = {}
    for old_label, new_label in label_mapping.items():
        if new_label not in obtained_tracklets:
            obtained_tracklets[new_label] = []
        obtained_tracklets[new_label].append(old_label)
    # Convert to sets for easier comparison
    obtained_tracklet_sets = [set(ids) for ids in obtained_tracklets.values()]
    expected_tracklet_sets = [set(ids) for ids in expected_tracklets.values()]

    # Check that each expected tracklet matches exactly one obtained tracklet
    for exp_name, exp_ids in expected_tracklets.items():
        exp_set = set(exp_ids)
        matching_obtained = [obt_set for obt_set in obtained_tracklet_sets if obt_set == exp_set]
        assert len(matching_obtained) == 1, f"Expected tracklet '{exp_name}' with IDs {exp_ids} was not found exactly once. Found {len(matching_obtained)} matches."

    # Check that no two expected tracklets are merged into one
    for obt_label, obt_ids in obtained_tracklets.items():
        obt_set = set(obt_ids)
        # Count how many expected tracklets have all their IDs in this obtained tracklet
        contain_expected = [exp_name for exp_name, exp_ids in expected_tracklets.items() if set(exp_ids).issubset(obt_set)]
        assert len(contain_expected) <= 1, f"Obtained tracklet with label {obt_label} contains IDs from multiple expected tracklets: {contain_expected}. IDs: {obt_ids}"

def test_import_trackmate():
    """ Test import a Trackmate .xml tracks into EpiCure structure """
    raw_path = os.path.join( ".", "test_data", "013_crop.tif" )
    xml_path = os.path.join(".", "test_data", "epics", "013_crop.xml")

    epic = epicure.EpiCure()
    epic.verbose = 3  # 0: minimal to 3: debug informations
    epic.load_movie(raw_path)
    epic.go_epicure("epics", xml_path)
    ## should be tracked
    assert epic.tracked>0
    ## should have loaded divisions from the trackmate graph
    assert epic.nb_divisions()>0


if __name__ == "__main__":
    test_import_trackmate()
    print("********* Test import/export to TrackMate XML completed ***********")


import numpy as np
import os
import epicure.epicuring as epi

def test_track_methods():
    """ Tracking with EpiCure with Laptrack, different parameters """
    test_img = os.path.join(".", "test_data", "003_crop.tif")
    test_seg = os.path.join(".", "test_data", "003_crop_epyseg.tif")

    ## load and initialize
    epic = epi.EpiCure()
    epic.load_movie(test_img)
    epic.go_epicure("test_epics", test_seg)

    ## no tracked yet 
    assert epic.tracked == 0
    alllabels = epic.nlabels()
    assert alllabels == 1294

    # default tracking
    track = epic.tracking
    ## Chosse method overlap for tracking
    track.track_choice.setCurrentText("Laptrack-Overlaps")
    track.min_iou.setText("0.1")
    track.split_cost.setText("0.5")
    track.merg_cost.setText("0")
    track.do_tracking()
    ## after tracking should have much less labels now
    ntracks = epic.nlabels()
    assert ntracks < alllabels 
    assert ntracks == track.nb_tracks()
    assert track.graph is not None
    ## found divisions 
    ## check two first results in graph are the daugthers: mother
    first = list(track.graph.keys())[0]
    first_mother = track.graph[first]
    second = list(track.graph.keys())[1]
    assert track.graph[second] == first_mother
    assert not track.check_gap()

    ## check reset function: reread the tracks from the labels (so should be the same)
    track.reset_tracks()
    assert epic.nlabels() == ntracks 
    assert epic.nlabels() == track.nb_tracks()

    ## check one track validity
    ## first tracking, no modification so should be ordered
    track_id = track.get_track_list()[20]
    assert track_id == 21
    assert track.get_first_frame( track_id ) == 0
    feats = track.measure_track_features( track_id )
    assert feats["TrackDuration"] == 11
    assert feats["NbGaps"] == 0 
    assert feats["Label"] == track_id 

    # Operations to get last frame, remove it
    last = track.get_last_frame(track_id)
    track.remove_one_frame( track_id, last )
    newlast = track.get_last_frame(track_id)
    assert last == newlast+1

    ## create a gap in the middle of the track, then fix it (split the track)
    midle = track.get_first_frame(track_id) + 2
    track.remove_one_frame( track_id, midle )
    gaped = track.check_gap()
    ## gaps are allowed now
    assert len(gaped) > 0
    epic.handle_gaps( None )
    assert track.nb_tracks() == ntracks + 1 

if __name__ == "__main__":
    test_track_methods()
    print("********* Test tracking cure completed ***********")


import numpy as np
import os
import epicure.epicuring as epi
import epicure.Utils as ut
#from unittest.mock import Mock
#from vispy import keys
import napari


def test_suspect_frame():
    """ Flag possible errors on a single frame, looking at possible outliers """
    test_img = os.path.join(".", "test_data", "area3_Composite.tif")
    test_seg = os.path.join(".", "test_data", "area3_Composite_epyseg.tif")
    epic = epi.EpiCure()
    epic.load_movie(test_img)
    epic.set_chanel(1, 1)
    assert epic.viewer is not None
    epic.go_epicure("test_epics", test_seg)

    ## flag cells if area smaller than given threshold 
    segedit = epic.inspecting
    assert segedit is not None
    segedit.min_area.setText("70")
    segedit.event_area_threshold()
    assert "Events" in epic.viewer.layers
    outlier = epic.viewer.layers["Events"]
    assert len(outlier.data)>=5
    ## check is done on frame 0 only
    assert outlier.data[1][0] == 0

    ## flag if the intensity ratio is to high 
    nsus = len(outlier.data)
    segedit.fintensity_out.setText("0.5")
    segedit.event_intensity(True)
    assert len(outlier.data) > (nsus+5)

def test_events_division():
    """ Handling events: add/remove division event """
    test_img = os.path.join(".", "test_data", "013_crop.tif")
    ## load and initialize
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    epic.load_movie(test_img)
    epic.go_epicure("epics")
    susp = epic.inspecting

    ## initialisation: only one division
    assert susp.nb_events() == 1
    ## check it's a division
    ind = 0
    assert susp.is_division(ind)
    assert len(epic.tracking.graph) == 2

    ## remove this event
    susp.exonerate_one(ind, remove_division=True)
    assert susp.nb_events() == 0
    ## check it was removed from the tracks graph
    assert len(epic.tracking.graph) == 0

    ## add it back
    epic.editing.add_division(490,477,3) ## added in Edit part
    assert susp.is_division(ind)
    print(epic.tracking.graph)
    assert len(epic.tracking.graph) == 2
    ## check that parent found was the correct one (73)
    assert epic.tracking.graph[490] == [73]

def test_events_extrusion():
    """ Handling events: detect/remove extrusion event """
    test_img = os.path.join(".", "test_data", "013_crop.tif")
    ## load and initialize
    viewer = napari.Viewer(show=False)
    epic = epi.EpiCure(viewer)
    epic.load_movie(test_img)
    epic.go_epicure("epics")
    susp = epic.inspecting

    ## initialisation: only one division
    assert susp.nb_events() == 1
    ## check it's a division
    ind = 0
    assert susp.is_division(ind)
    assert len(epic.tracking.graph) == 2

    ## Parameters for track inspection: only get extrusions
    ## disable all options
    susp.ignore_borders.setChecked(False)
    susp.ignore_boundaries.setChecked(False)
    susp.check_jump.setChecked(False)
    susp.check_length.setChecked(False)
    susp.check_shape.setChecked(False)
    susp.check_size.setChecked(False)
    susp.get_apparition.setChecked(False)
    susp.get_disparition.setChecked(False)
    susp.get_division.setChecked(False)
    susp.get_gaps.setChecked(False)
    ## active only get extrusions option 
    susp.threshold_disparition.setText("100")
    susp.get_extrusions.setChecked(True)

    ## do the inspection
    susp.inspect_tracks(subprogress=False)
    ## should have found one extrusion
    assert susp.nb_events() == 2
    assert susp.is_extrusion(1)

    ## remove it
    ## select it as current event
    susp.event_num.setValue(1)
    ## zoom on it
    susp.go_to_event()
    ## remove the event
    susp.clear_event()
    assert susp.nb_events() == 1
    assert susp.is_division(ind)


def test_suspect_track():
    """ Track and flag weird tracks """
    test_img = os.path.join(".", "test_data", "003_crop.tif")
    test_seg = os.path.join(".", "test_data", "003_crop_epyseg.tif")

    ## load and initialize
    epic = epi.EpiCure()
    epic.load_movie(test_img)
    epic.go_epicure("test_epics", test_seg)
    track = epic.tracking
    susp = epic.inspecting

    ## initialisation: no suspects
    assert susp.nb_events() == 0
    # default tracking
    track.do_tracking()
    ## after tracking, only possible events are division 
    assert susp.nb_events() == susp.nb_type("division")
    nev = susp.nb_events()
    assert nev > 0
    ## manually add a suspect
    susp.add_event( (5,50,50), 10, "test" )
    assert susp.nb_events() == (nev+1)
    ## test default parameter inspection
    susp.check_size.setChecked( False )
    susp.check_length.setChecked( False )
    susp.inspect_tracks()
    assert susp.nb_events() > 30
    assert susp.nb_events() < 60
    ## test minimum track length inspection
    susp.check_length.setChecked( True )
    susp.min_length.setText("5")
    nmin_prev =  susp.nb_events()
    susp.inspect_tracks()
    nmin =  susp.nb_events()
    assert nmin > nmin_prev
    assert nmin > 50 
    ## test reset all
    susp.reset_all_events()
    assert susp.nb_events() == 0
    ## Test reloading the divisions from the track graph
    susp.get_divisions()
    assert susp.nb_events() == nev
    ## Track feature change test
    susp.check_size.setChecked( True )
    susp.inspect_tracks()
    assert susp.nb_events() > 50 

def test_boundaries():
    """ Detecting cells on border/boundaries and removing border cells """
    test_img = os.path.join(".", "test_data", "003_crop.tif")
    test_seg = os.path.join(".", "test_data", "003_crop_epyseg.tif")

    ## load and initialize
    epic = epi.EpiCure()
    epic.load_movie(test_img)
    epic.go_epicure("test_epics", test_seg)

    ## check that doesn't find any boundary cells (touching background)
    epic.inspecting.get_boundaries_cells()
    assert len(epic.inspecting.boundary_cells[0]) == 0

    ## remove the border cells, so now should find boundary cells
    epic.editing.remove_border()
    epic.inspecting.get_boundaries_cells()
    assert len(epic.inspecting.boundary_cells[0]) > 20

if __name__ == "__main__":
    #test_suspect_frame()
    #test_suspect_track()
    #test_boundaries()
    #test_events_division()
    test_events_extrusion()
    print("********* Test suspects cure completed ***********")



"""
Module for loading TrackMate files into EpiCure.

This loader uses an iterative parsing approach to efficiently handle large XML files.

Relevant metadata such as time step and pixel size are stored into a dictionary.
The cells positions are stored in a unique NumPy array with columns for label, time,
x position, and y position.
Segmentations are stored as NumPy arrays with dimensions matching the original image data.
Tracks are stored as a dictionary mapping daughter cell labels to their mother cell labels:
{label_of_daughter_cell: [label_of_mother_cell]}
"""

import xml.etree.ElementTree as ET
from copy import deepcopy
from pathlib import Path
from typing import Iterator, Union

import numpy as np
from skimage.draw import polygon2mask
import epicure.Utils as ut


def _get_ImageData_tag(xml_path: Path) -> ET.Element:
    """
    Extract the 'ImageData' tag from an XML file.

    This function parses an XML file to find and extract the 'ImageData' tag.
    Once found, a new element with only the attributes is created and returned.

    Args:
        xml_path (Path): The file path of the XML file to be parsed.

    Returns:
        ET.Element: An `ET.Element` object with the 'ImageData' attributes.

    Raises:
        LookupError: If the 'ImageData' tag is not found in the XML file.
    """
    img_data_tag = None
    with open(xml_path, "rb") as f:
        it = ET.iterparse(f, events=["start", "end"])
        _, root = next(it)  # Saving the root of the tree for later cleaning.

        for event, element in it:
            if event == "end" and element.tag == "ImageData":
                # Create a new element with only the attributes (no children subtree).
                img_data_tag = ET.Element(element.tag, element.attrib)
                root.clear()  # Cleaning the tree to free up memory.
                break  # We found what we need, exit early.
            elif event == "end":
                element.clear()

    if img_data_tag is None:
        raise LookupError("The 'ImageData' tag was not found in the XML file.")

    return img_data_tag


def _get_metadata(img_data: ET.Element) -> dict[str, Union[int, float, str]]:
    """
    Extract metadata from the 'ImageData' XML element.

    Parameters
    ----------
    img_data : ET.Element
        The XML element containing the 'ImageData' information.

    Returns
    -------
    dict
        A dictionary containing the extracted image metadata.
    """
    int_keys = ["width", "height", "nframes"]
    float_keys = ["pixelwidth", "pixelheight", "timeinterval"]
    metadata = {}
    for key in int_keys + float_keys:
        metadata[key] = img_data.attrib.get(key)

    for key in int_keys:
        if metadata[key] is None:
            raise KeyError(f"No '{key}' attribute in the 'ImageData' XML element.")
        metadata[key] = int(metadata[key])
    for key in float_keys:
        if metadata[key] is None:
            raise KeyError(f"No '{key}' attribute in the 'ImageData' XML element.")
        metadata[key] = float(metadata[key])

    # TODO: check if non-square pixels ok.
    assert metadata["pixelwidth"] == metadata["pixelheight"], "Non-square pixels. Unsure if supported?"

    return metadata


def _get_units(
    element: ET.Element,
) -> dict[str, str]:
    """Extract units information from an XML element and return it as a dictionary.

    This function deep copies the attributes of the XML element into a dictionary,
    then clears the element to free up memory.

    Args:
        element (ET._Element): The XML element holding the units information.

    Returns:
        dict[str, str]: A dictionary containing the units information.
        Keys are 'spatialunits' and 'timeunits'.

    Warns:
        If the 'spatialunits' or 'timeunits' attributes are not found,
        defaulting them to 'pixel' and 'frame', respectively.
    """
    units = {}  # type: dict[str, str]
    if element.attrib:
        units = deepcopy(element.attrib)
    if "spatialunits" not in units:
        ut.show_warning("No space unit found in the XML file. Setting to 'pixel'.")
        units["spatialunits"] = "pixel"  # TrackMate default value
    if "timeunits" not in units:
        ut.show_warning("No time unit found in the XML file. Setting to 'frame'.")
        units["timeunits"] = "frame"  # TrackMate default value
    element.clear()  # We won't need it anymore so we free up some memory.
    # .clear() does not delete the element: it only removes all subelements
    # and clears or sets to `None` all attributes.
    return units


def _parse_all_spots(
    it: Iterator[tuple[str, ET.Element]],
    positions: np.ndarray,
    segmentation: np.ndarray,
    metadata: dict[str, Union[int, float, str]],
) -> list[int]:
    """
    Parse the 'AllSpots' XML element to extract spot positions and segmentation data.

    This function iterates through the XML elements under 'AllSpots' to extract
    spot positions and update the segmentation array.

    Args:
        it (ET.iterparse): An iterator for parsing XML elements.
        positions (np.ndarray): A NumPy array to store the extracted positions.
        segmentation (np.ndarray): A NumPy array to store segmentation data.
        metadata (dict[str, Union[int, float, str]]): A dictionary containing units information.

    Returns:
        list[int]: A list of spot IDs filtered out in TrackMate.
    """
    px_width = float(metadata.get("pixelwidth", 1.0))
    px_height = float(metadata.get("pixelheight", 1.0))
    spot_index = 0
    invisible_spots_ids = []
    for event, element in it:
        if element.tag == "Spot" and event == "end":
            if not int(element.attrib["VISIBILITY"]):
                invisible_spots_ids.append(int(element.attrib["ID"]))
                element.clear()
                continue  # skip invisible spots

            t = int(float(element.attrib["FRAME"]))
            x = float(element.attrib["POSITION_X"]) / px_width
            y = float(element.attrib["POSITION_Y"]) / px_height
            label = int(element.attrib["ID"])
            positions[spot_index] = [label, t, x, y]

            contour = element.text
            npoints = int(element.attrib["ROI_N_POINTS"])

            if contour is not None:
                coords = np.array([float(x) for x in contour.split()])
                dimension = len(coords) // npoints
                coords = coords.reshape(-1, dimension)
                coords[:, 0] = x + (coords[:, 0] / px_width)
                coords[:, 1] = y + (coords[:, 1] / px_height)
                contour_rc = np.flip(coords, axis=1)  # x, y to row, col
                mask = polygon2mask(segmentation[t].shape, contour_rc)
                segmentation[t][mask] = label
            else:
                print(f"no contour for spot {label}, frame {t}")

            spot_index += 1
            element.clear()
        elif element.tag == "AllSpots" and event == "end":
            break

    return invisible_spots_ids


def _parse_all_tracks(it: Iterator[tuple[str, ET.Element]], tracks: dict[int, list[int]]) -> None:
    """
    Parse the 'AllTracks' XML element to extract track information.

    This function iterates through the XML elements under 'AllTracks' to extract
    track information and populate the tracks dictionary. This dictionary maps
    daughter cell labels to their mother cell labels.

    Args:
        it (ET.iterparse): An iterator for parsing XML elements.
        tracks (dict[int, list[int]]): A dictionary to store the extracted tracks.
    """
    for event, element in it:
        if element.tag == "Edge" and event == "start":
            mother_id = int(element.attrib["SPOT_SOURCE_ID"])
            daughter_id = int(element.attrib["SPOT_TARGET_ID"])
            if daughter_id not in tracks:
                tracks[daughter_id] = [mother_id]
            else:
                tracks[daughter_id].append(mother_id)
            element.clear()

        elif element.tag == "AllTracks" and event == "end":
            break


def _build_label_mapping(positions: np.ndarray, tracks: dict[int, list[int]]) -> dict[int, int]:
    """
    Build a mapping from TrackMate labels to EpiCure labels.

    In TrackMate, each detected spot has a unique label, while in EpiCure,
    labels are constant per tracklet, hence the need for mapping.

    Args:
        positions (np.ndarray): The array of positions.
        tracks (dict[int, list[int]]): The dictionary of tracks.

    Returns:
        dict[int, int]: A dictionary mapping TrackMate labels to EpiCure labels.
    """
    # Reverse mapping from daughter to mother to get simple edges and division edges.
    mother_to_daughters = {}
    for daughter, mothers in tracks.items():
        for mother in mothers:
            if mother not in mother_to_daughters:
                mother_to_daughters[mother] = [daughter]
            else:
                mother_to_daughters[mother].append(daughter)

    edges = {m: d[0] for m, d in mother_to_daughters.items() if len(d) == 1}
    divisions = {m: d for m, d in mother_to_daughters.items() if len(d) > 1}
    fusions = {d: m for d, m in tracks.items() if len(m) > 1}
    fusion_mothers = set()  # a set of mothers that participate in fusions
    for mothers in fusions.values():
        fusion_mothers.update(mothers)

    label_mapping = {}
    new_label = 1
    frames = np.unique(positions[:, 1])

    for frame in frames:
        frame_positions = positions[positions[:, 1] == frame]
        for old_label in frame_positions[:, 0]:
            old_label_int = int(old_label)

            # Division edge => mother keeps its label, each daughter gets a new label.
            if old_label_int in divisions:
                # Assign new label to the mother when not already assigned (tracklet start).
                if old_label_int not in label_mapping:
                    label_mapping[old_label_int] = new_label
                    new_label += 1
                # Assign new labels to each daughter.
                daughters = divisions[old_label_int]
                for daughter in daughters:
                    label_mapping[int(daughter)] = new_label
                    new_label += 1

            # Mother in a fusion => gets its own label, tracklet ends here.
            elif old_label_int in fusion_mothers:
                if old_label_int not in label_mapping:
                    label_mapping[old_label_int] = new_label
                    new_label += 1

            # Daughter in a fusion => gets a new label (multiple mothers merge into this).
            elif old_label_int in fusions:
                if old_label_int not in label_mapping:
                    label_mapping[old_label_int] = new_label
                    new_label += 1
                # If this fusion daughter is also a mother in a simple edge, propagate the label.
                if old_label_int in edges:
                    daughter_id = int(edges[old_label_int])
                    if daughter_id not in label_mapping:
                        label_mapping[daughter_id] = label_mapping[old_label_int]

            # Simple edge => mother and daughter share the same label (same tracklet).
            elif old_label_int in edges:
                if old_label_int not in label_mapping:
                    label_mapping[old_label_int] = new_label
                    new_label += 1
                # Propagate the same label to the daughter (continue tracklet).
                daughter_id = int(edges[old_label_int])
                if daughter_id not in label_mapping:
                    label_mapping[daughter_id] = label_mapping[old_label_int]

            # Lone detection or start of new track.
            elif old_label_int not in label_mapping:
                label_mapping[old_label_int] = new_label
                new_label += 1

    # Do we have everyone mapped?
    assert len(label_mapping) == positions.shape[0], "Some labels were not mapped!"
    assert sorted(label_mapping.keys()) == sorted(set(positions[:, 0].astype(int))), "Some labels were not mapped!"

    return label_mapping


def relabel_positions(label_mapping: dict[int, int], positions: np.ndarray) -> np.ndarray:
    """
    Relabel positions to match EpiCure requirements.

    Args:
        label_mapping (dict[int, int]): A dictionary mapping TrackMate labels to EpiCure labels.
        positions (np.ndarray): The array of positions to be relabeled.

    Returns:
        np.ndarray: The relabeled positions.
    """
    new_positions = np.zeros_like(positions)
    for i in range(positions.shape[0]):
        old_label = int(positions[i, 0])
        new_label = label_mapping[old_label]
        new_positions[i] = positions[i]
        new_positions[i, 0] = new_label
    return new_positions


def relabel_tracks(label_mapping: dict[int, int], tracks: dict[int, list[int]]) -> dict[int, list[int]]:
    """
    Relabel tracks to match EpiCure requirements.

    Args:
        label_mapping (dict[int, int]): A dictionary mapping TrackMate labels to EpiCure labels.
        tracks (dict[int, list[int]]): The dictionary of tracks to be relabeled.

    Returns:
        dict[int, list[int]]: The relabeled tracks.
    """
    new_tracks = {}
    for daughter_old, mothers_old in tracks.items():
        daughter_new = label_mapping[daughter_old]
        mothers_new = [label_mapping[mother_old] for mother_old in mothers_old]
        # Ignore entries for which the daughter label is identical to the mother(s) label.
        if daughter_new not in mothers_new:
            new_tracks[daughter_new] = mothers_new
    return new_tracks


def relabel_segmentation(label_mapping: dict[int, int], segmentation: np.ndarray) -> np.ndarray:
    """
    Relabel segmentation to match EpiCure requirements.

    Args:
        label_mapping (dict[int, int]): A dictionary mapping TrackMate labels to EpiCure labels.
        segmentation (np.ndarray): The segmentation array to be relabeled.

    Returns:
        np.ndarray: The relabeled segmentation.
    """
    new_seg = np.zeros_like(segmentation)
    for old_label, new_label in label_mapping.items():
        new_seg[segmentation == old_label] = new_label
    return new_seg


def _parse_Model_tag(
    xml_path: Path,
    metadata: dict[str, Union[int, float, str]],
    segmentation: np.ndarray,
) -> tuple[np.ndarray, dict[int, list[int]]]:
    """
    Extract the 'Model' tag from an XML file.

    This function parses an XML file to find and extract the 'Model' tag.
    Once found, the tag is deep copied and returned.

    Args:
        xml_path (Path): The file path of the XML file to be parsed.
        metadata (dict[str, int | float]): A dictionary to update with extracted units information.
        segmentation (np.ndarray): A NumPy array to store segmentation data.

    Returns:
        np.ndarray: A NumPy array containing the positions data.
        dict[int, list[int]]: A dictionary containing the tracks data.
    """
    ignored_spots = None
    with open(xml_path, "rb") as f:
        it = ET.iterparse(f, events=["start", "end"])
        _, root = next(it)  # saving the root of the tree for later cleaning

        units: dict[str, str] = {}
        positions: np.ndarray = np.empty((0, 4), dtype=np.float32)
        tracks: dict[int, list[int]] = {}
        for event, element in it:
            # Check for the 'Model' tag
            if element.tag == "Model" and event == "start":
                units = _get_units(element)
                metadata.update(units)
                root.clear()  # cleaning the tree to free up some memory
                # All the browsed subelements of `root` are deleted.

            # From AllSpots we extract the positions and segmentation.
            if element.tag == "AllSpots" and event == "start":
                positions = np.zeros((int(element.attrib["nspots"]), 4), dtype=np.float32)
                ignored_spots = _parse_all_spots(it, positions, segmentation, metadata)
                root.clear()

            # From AllTracks we extract the dict of tracks.
            if element.tag == "AllTracks" and event == "start":
                _parse_all_tracks(it, tracks)
                root.clear()

            if element.tag == "Model" and event == "end":
                root.clear()
                break  # not interested in the following data

    if ignored_spots is not None:
        ut.show_warning(f"{len(ignored_spots)} spots were filtered out in TrackMate and will not be loaded into EpiCure. IDs: {ignored_spots}.")
        # The array positions was initialized with the total number of spots,
        # but since some spots were ignored, we need to filter them out.
        positions = positions[: positions.shape[0] - len(ignored_spots)]

    return positions, tracks




import numpy as np
import os
import epicure.epicuring as epi
import napari
import epicure.Utils as ut
from epicure.start_epicuring import gui_files


def test_load_movie():
    """ Read a standard tif movie """
    test_img = os.path.join(".", "test_data", "003_crop.tif")
    epic = epi.EpiCure()
    epic.load_movie(test_img)
    assert epic.img.shape == (11,189,212)
    return

def test_load_image():
    """ Read a single image and a cellpose (labels) segmentation """
    test_img = os.path.join(".", "test_data", "static_image.tif")
    test_seg = os.path.join(".", "test_data", "static_image_cellpose.tif")
    epic = epi.EpiCure()
    epic.load_movie(test_img)
    assert epic.img.shape == (1,507,585)
    epic.load_segmentation(test_seg)
    assert epic.seg.shape == epic.img.shape
    assert np.max(epic.seg) == 706
    return

def test_load_movie_with_chanel():
    """ Read a tif movie with 2 channels """
    test_img = os.path.join(".", "test_data", "area3_Composite.tif")
    test_seg = os.path.join(".", "test_data", "area3_Composite_epyseg.tif")
    epic = epi.EpiCure()
    resaxis, resval = epic.load_movie(test_img)
    # check the dimensions are correctly loaded
    assert epic.img.shape == (5,2,146,228)
    assert resaxis == 1
    assert resval == 2
    ## check the channels are correctly set
    assert np.mean(epic.img)>=150
    epic.set_chanel(1, 1)
    assert np.mean(epic.img)<150
    assert np.mean(epic.img)>100
    return

def test_load_segmentation():
    test_seg = os.path.join(".", "test_data", "003_crop_epyseg.tif")
    epic = epi.EpiCure()
    epic.load_segmentation(test_seg)
    assert epic.seg.shape == (11,189,212)
    assert np.max(epic.seg) == 1294
    return

def test_suggest():
    """ Check segmentation file name suggestion """
    ## case 1: file doesn't exists, creates it
    test_img = os.path.join(".", "test_data", "003_crop.tif")
    epic = epi.EpiCure()
    epic.load_movie(test_img)
    segfile = epic.suggest_segfile("epics")
    assert segfile is None
    ## case 2: if it exists, find it automatically
    test_img = os.path.join(".", "test_data", "013_crop.tif")
    epic.load_movie(test_img)
    segfile = epic.suggest_segfile("epics")
    resfile = os.path.join(".", "test_data", "epics", "013_crop_labels.tif")
    absp = os.path.abspath(resfile)
    assert segfile == absp 
    return

def test_init_epic():
    epic = epi.EpiCure()
    assert epic.img is None
    return

def test_load_from_layers():
    """ Open a new EpiCure project from opened layers """
    viewer = napari.Viewer(show=False)
    test_img = os.path.join(".", "test_data", "003_crop.tif")
    img, _, _, _, _, _ = ut.open_image(test_img, get_metadata=False, verbose=0)
    movie_layer = viewer.add_image(img, name="TestImage")
    ## if initiliaze without segmentation, propose option to choose/do segmentation in the interface
    widget, epic = gui_files(movie_layer, test_img, None)
    assert (os.path.abspath(widget.segmentation_file.value)==os.path.abspath(os.path.join(".", "test_data")))
    ## loading of movie has worked
    assert "Movie" in viewer.layers
    ## load also segmentation from open layers
    ## reset
    ut.remove_layer(viewer, "Movie")
    movie_layer = viewer.add_image(img, name="TestImage")
    test_seg = os.path.join(".", "test_data", "003_crop_epyseg.tif")
    seg, _, _, _, _, _ = ut.open_image(test_seg, get_metadata=False, verbose=0)
    seg_layer = viewer.add_labels(seg, name="TestSegmentation")
    widget, epic = gui_files(movie_layer, test_img, seg_layer)
    assert "Movie" in viewer.layers
    ## launch epicure with the given layers
    widget.call_button.clicked(False)
    ## check that it was properly loaded
    assert "Segmentation" in viewer.layers
    assert epic.imgshape2D == img.shape[1:]
    assert epic.nframes == img.shape[0]
    assert epic.nlabels() == 1294
    assert os.path.abspath(epic.outdir) == os.path.abspath(os.path.join(".", "test_data", "epics"))
    #viewer.show() # manual check
    
if __name__ == "__main__":
    test_load_from_layers()
    test_load_image()
    print("********* Test basics completed ***********")
